module FlexFIR where

import CoreLoom

fir :: Datapath
fir = makeDatapath dp
 where
  dp = do
     inputs ["Data_in", "K_0", "K_1", "K_2", "K_3"]
     outputs ["Data_out"]
     functions MUL ["M_0", "M_1", "M_2", "M_3"]
     functions ADD ["A_1", "A_2", "A_3"]
     delayN 6 open "D"
     delays open ["S_0", "S_1", "S_2"]
     connect ("Data_in", "D_5")
     connections_n [("Data_in", ("M_0", 1)), ("D_4", ("M_1", 1))]
     connections_n [("D_2", ("M_2", 1)), ("D_0", ("M_3", 1))]
     connections_n [("K_0", ("M_0", 0)), ("K_1", ("M_1", 0))]
     connections_n [("K_2", ("M_2", 0)), ("K_3", ("M_3", 0))]
     connections_n [("M_0", ("S_0", 0)), ("M_1", ("A_1", 1))]
     connections_n [("M_2", ("A_2", 1)), ("M_3", ("A_3", 1))]
     connections_n [("S_0", ("A_1", 0)), ("S_1", ("A_2", 0))]
     connections_n [("S_2", ("A_3", 0))]
     connections [("A_1", "S_1"), ("A_2", "S_2"), ("A_3", "Data_out")]


flexFIR :: Datapath
flexFIR = makeDatapath dp
 where
  dp = do
     inputs ["Data_in", "K_0", "K_1", "K_2", "K_3"]
     inputs ["V_in_0", "V_in_1", "V_in_2", "V_in_3"]
     outputs ["Data_out"]
     outputs ["V_out_0", "V_out_1", "V_out_2", "V_out_3"]
     functions MUL ["M_0", "M_1", "M_2", "M_3"]
     functions ADD ["A_0", "A_1", "A_2", "A_3"]
     constant (VInt 0) "Zero"
     delayN 6 open "D"
     delays open ["S_0", "S_1", "S_2"]
     connect ("Data_in", "D_5")
     connections_n [("Data_in", ("M_0", 1)), ("D_4", ("M_1", 1))]
     connections_n [("D_2", ("M_2", 1)), ("D_0", ("M_3", 1))]
     connections_n [("K_0", ("M_0", 0)), ("K_1", ("M_1", 0))]
     connections_n [("K_2", ("M_2", 0)), ("K_3", ("M_3", 0))]
     connections_n [("K_0", ("A_0", 0)), ("K_1", ("A_1", 0))]
     connections_n [("K_2", ("A_2", 0)), ("K_3", ("A_3", 0))]
     connections_n [("Data_in", ("M_1", 1)), ("Data_in", ("M_2", 1))]
     connections_n [("Data_in", ("M_3", 1))]
     connections_n [("Data_in", ("A_0", 1)), ("Data_in", ("A_1", 1))]
     connections_n [("Data_in", ("A_2", 1)), ("Data_in", ("A_3", 1))]
     connections_n [("V_in_0", ("M_0", 1)), ("V_in_1", ("M_1", 1))]
     connections_n [("V_in_2", ("M_2", 1)), ("V_in_3", ("M_3", 1))]
     connections_n [("M_0", ("A_0", 1)), ("M_1", ("A_1", 1))]
     connections_n [("M_2", ("A_2", 1)), ("M_3", ("A_3", 1))]
     connections_n [("V_in_0", ("A_0", 1)), ("V_in_1", ("A_1", 1))]
     connections_n [("V_in_2", ("A_2", 1)), ("V_in_3", ("A_3", 1))]
     connections_n [("Zero", ("A_0", 0)), ("S_0", ("A_1", 0))]
     connections_n [("S_1", ("A_2", 0)), ("S_2", ("A_3", 0))]
     connections [("A_0", "S_0"), ("A_1", "S_1"), ("A_2", "S_2"), ("A_3", "Data_out")]
     connections [("M_0", "V_out_0"), ("M_1", "V_out_1")]
     connections [("M_2", "V_out_2"), ("M_3", "V_out_3")]
     connections [("A_0", "V_out_0"), ("A_1", "V_out_1")]
     connections [("A_2", "V_out_2"), ("A_3", "V_out_3")]

cf_FIR :: Config
cf_FIR = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
           "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
           "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
           "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)"],
          [[Var "Data_in",  Var "D_4",      Var "D_2",      Var "D_0",
	    Var "Zero",     Var "S_0",      Var "S_1",      Var "S_2",
            Var "M_0",      Var "M_1",      Var "M_2",      Var "M_3",
	    open,           open,           open,           open]])

cf_VSMUL :: Config
cf_VSMUL = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
             "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
             "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
             "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)"],
            [[Var "Data_in",  Var "Data_in",   Var "Data_in", Var "Data_in",
              open,           open,            open,          open,
              open,           open,            open,          open,
	      Var "M_0",      Var "M_1",       Var "M_2",     Var "M_3"]])

cf_VSADD :: Config
cf_VSADD = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
             "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
             "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
             "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)"],
            [[open,           open,           open,           open,
	      Var "K_0",      Var "K_1",      Var "K_2",      Var "K_3",
	      Var "Data_in",  Var "Data_in",  Var "Data_in",  Var "Data_in",
	      Var "A_0",      Var "A_1",      Var "A_2",      Var "A_3"]])	  

cf_VVMUL :: Config
cf_VVMUL = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
             "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
             "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
             "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)"],
            [[Var "V_in_0",   Var "V_in_1",   Var "V_in_2",   Var "V_in_3",
              open,           open,           open,           open,
              open,           open,           open,           open,
	      Var "M_0",      Var "M_1",      Var "M_2",      Var "M_3"]])

cf_VVADD :: Config
cf_VVADD = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
             "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
             "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
             "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)"],
            [[open,           open,           open,           open,
	      Var "K_0",      Var "K_1",      Var "K_2",      Var "K_3",
	      Var "V_in_0",   Var "V_in_1",   Var "V_in_2",   Var "V_in_3",
	      Var "A_0",      Var "A_1",      Var "A_2",      Var "A_3"]])	

cf_FIR3 :: Config
cf_FIR3 = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
            "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
            "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
            "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)"],
           [[Var "Data_in",  Var "D_4",      Var "D_2",      open,
	     Var "Zero",     Var "S_0",      Var "S_1",      open,
             Var "M_0",      Var "M_1",      Var "M_2",      open,
	     open,           open,           Var "A_2",      open]])

flexFIR_DP :: Datapath
flexFIR_DP = makeDatapath dp
 where
  dp = do
     inputs ["Data_in", "K_0", "K_1", "K_2", "K_3"]
     inputs ["V_in_0", "V_in_1", "V_in_2", "V_in_3"]
     outputs ["Data_out"]
     outputs ["V_out_0", "V_out_1", "V_out_2", "V_out_3"]
     functions MUL ["M_0", "M_1", "M_2", "M_3"]
     functions ADD ["A_0", "A_1", "A_2", "A_3"]
     constant (VInt 0) "Zero"
     delayN 6 open "D"
     delay open "B_0"
     delayN 2 open "C"
     delays open ["S_0", "S_1", "S_2"]
     connect ("Data_in", "D_5")
     connections_n [("Data_in", ("M_0", 1)), ("D_4", ("M_1", 1))]
     connections_n [("D_2", ("M_2", 1)), ("D_0", ("M_3", 1))]
     connections_n [("K_0", ("M_0", 0)), ("K_1", ("M_1", 0))]
     connections_n [("K_2", ("M_2", 0)), ("K_3", ("M_3", 0))]
     connections [("K_1", "B_0"), ("K_2", "C_1")]
     connections_n [("B_0", ("M_1", 0)), ("C_0", ("M_2", 0))] 
     connections_n [("K_0", ("A_0", 0)), ("K_1", ("A_1", 0))]
     connections_n [("K_2", ("A_2", 0)), ("K_3", ("A_3", 0))]
     connections_n [("Data_in", ("M_1", 1)), ("Data_in", ("M_2", 1))]
     connections_n [("Data_in", ("M_3", 1))]
     connections_n [("Data_in", ("A_0", 1)), ("Data_in", ("A_1", 1))]
     connections_n [("Data_in", ("A_2", 1)), ("Data_in", ("A_3", 1))]
     connections_n [("V_in_0", ("M_0", 1)), ("V_in_1", ("M_1", 1))]
     connections_n [("V_in_2", ("M_2", 1)), ("V_in_3", ("M_3", 1))]
     connections_n [("M_0", ("A_0", 1)), ("M_1", ("A_1", 1))]
     connections_n [("M_2", ("A_2", 1)), ("M_3", ("A_3", 1))]
     connections_n [("V_in_0", ("A_0", 1)), ("V_in_1", ("A_1", 1))]
     connections_n [("V_in_2", ("A_2", 1)), ("V_in_3", ("A_3", 1))]
     connections [("V_in_1", "D_4"), ("V_in_2", "D_3")]
     connections_n [("Zero", ("A_0", 0)), ("S_0", ("A_1", 0))]
     connections_n [("S_1", ("A_2", 0)), ("S_2", ("A_3", 0))]
     connections [("A_0", "S_0"), ("A_1", "S_1")]
     connections [("A_2", "S_2"), ("A_3", "Data_out")]
     connections [("M_0", "V_out_0"), ("M_1", "V_out_1")]
     connections [("M_2", "V_out_2"), ("M_3", "V_out_3")]
     connections [("A_0", "V_out_0"), ("A_1", "V_out_1")]
     connections [("A_2", "V_out_2"), ("A_3", "V_out_3")]

cf_DP_FIR :: Config
cf_DP_FIR = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
              "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
              "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
              "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)",
	      "into_(D_4,0)", "into_(D_3,0)", "into_(M_1,0)", "into_(M_2,0)"],
            [[Var "Data_in",  Var "D_4",       Var "D_2",      Var "D_0",
	      Var "Zero",     Var "S_0",       Var "S_1",      Var "S_2",
              Var "M_0",      Var "M_1",       Var "M_2",      Var "M_3",
	      open,           open,            open,           open,
	      Var "D_5",      Var "D_4",       Var "K_1",      Var "K_2"]])

cf_DP_DPROD :: Config
cf_DP_DPROD = (["into_(M_0,1)", "into_(M_1,1)", "into_(M_2,1)", "into_(M_3,1)",
                "into_(A_0,0)", "into_(A_1,0)", "into_(A_2,0)", "into_(A_3,0)",
                "into_(A_0,1)", "into_(A_1,1)", "into_(A_2,1)", "into_(A_3,1)",
                "into_(V_out_0,0)", "into_(V_out_1,0)", "into_(V_out_2,0)", "into_(V_out_3,0)",
	        "into_(D_4,0)", "into_(D_3,0)", "into_(M_1,0)", "into_(M_2,0)"],
              [[Var "V_in_0",    Var "D_4",      Var "D_2",      open,
	        Var "Zero",      Var "S_0",      Var "S_1",      open,
                Var "M_0",       Var "M_1",      Var "M_2",      open,
	        open,            open,           Var "A_2",      open,
	        Var "V_in_1",    Var "V_in_2",   Var "B_0",      Var "C_0"]])	    

	      	          	
{-        OLD MODULAR FIR IMPLEMENTATION
flexFIRStartMod :: Datapath
flexFIRStartMod = makeDatapath dp
 where
   dp = do
     inputs ["Data_in", "K", "Vector_in"]
     outputs ["Data_out", "Sum_out", "Mul_out", "Add_out"]
     function MUL "FU_M"
     function ADD "FU_A"
     constant (VInt 0) "Zero"
     connections_n [("Zero",("FU_A", 0)),("FU_M", ("FU_A", 1))]
     connections_n [("Data_in", ("FU_M", 1)), ("K", ("FU_M", 0))]
     connections [("Data_in", "Data_out"), ("FU_A", "Sum_out")]
     connections_n [("Vector_in", ("FU_M", 1)),("K", ("FU_A", 0))]
     connections_n [("Vector_in",("FU_A", 1))]
     connections [("FU_M", "Mul_out"), ("FU_A", "Add_out")]
     
flexFIRMod :: Datapath
flexFIRMod = makeDatapath dp
 where
   dp = do
     inputs ["Data_in", "K", "Sum_in", "Vector_in"]
     outputs ["Data_out", "Sum_out", "Mul_out", "Add_out"]
     function MUL "FU_M"
     function ADD "FU_A"
     delays open ["D1","D2", "D3"]
     connections [("Data_in", "D1"), ("D1", "D2"), ("Sum_in", "D3")]
     connections_n [("D2", ("FU_M", 1)), ("K", ("FU_M", 0))]
     connections_n [("D3", ("FU_A", 0)), ("FU_M", ("FU_A", 1))]
     connections [("D2", "Data_out"), ("FU_A", "Sum_out")]
     connections_n [("Vector_in", ("FU_M", 1)),("K", ("FU_A", 0))]
     connections_n [("Vector_in",("FU_A", 1))]
     connections [("FU_M", "Mul_out"), ("FU_A", "Add_out")]

flexFIR :: Datapath
flexFIR = makeDatapath dp
 where
  dp = do
    addDP flexFIRStartMod "FIR3"
    addDP flexFIRMod "FIR2"
    addDP flexFIRMod "FIR1"
    addDP flexFIRMod "FIR0"
    connections [("FIR3:Data_out", "FIR2:Data_in")]
    connections [("FIR2:Data_out", "FIR1:Data_in"),("FIR1:Data_out", "FIR0:Data_in")]
    connections [("FIR3:Sum_out","FIR2:Sum_in")]
    connections [("FIR2:Sum_out","FIR1:Sum_in"),("FIR1:Sum_out","FIR0:Sum_in")]
    remNode "FIR0:Data_out"

--FIR filter configuration
cf_flexFIRStartMod_FIR :: Config
cf_flexFIRStartMod_FIR =
 (["into_(FU_M,1)","into_(FU_A,0)","into_(FU_A,1)"],
  [[Var "Data_in",  Var "Zero",     Var "FU_M"]])
  
cf_flexFIRMod_FIR :: Config
cf_flexFIRMod_FIR =
 (["into_(FU_M,1)","into_(FU_A,0)","into_(FU_A,1)"],
  [[Var "D2",       Var "D3",       Var "FU_M"]])

cf_flexFIR_FIR :: Config
cf_flexFIR_FIR = concatConfig 
                   (subConfig cf_flexFIRStartMod_FIR "FIR3") 
		   (subConfigs cf_flexFIRMod_FIR ["FIR2", "FIR1", "FIR0"])

--Vector-Vector Multiply
cf_flexFIRStartMod_MUL :: Config
cf_flexFIRStartMod_MUL =
 (["into_(FU_M,1)","into_(FU_A,0)","into_(FU_A,1)"],
  [[Var "Vector_in", open,          open]])

cf_flexFIRMod_MUL :: Config
cf_flexFIRMod_MUL =
 (["into_(FU_M,1)","into_(FU_A,0)","into_(FU_A,1)"],
  [[Var "Vector_in", open,          open]])

cf_flexFIR_MUL :: Config
cf_flexFIR_MUL = concatConfig 
                   (subConfig cf_flexFIRStartMod_MUL "FIR3") 
		   (subConfigs cf_flexFIRMod_MUL ["FIR2", "FIR1", "FIR0"])

--Vector-Vector Addition
cf_flexFIRStartMod_ADD :: Config
cf_flexFIRStartMod_ADD =
 (["into_(FU_M,1)","into_(FU_A,0)", "into_(FU_A,1)"],
  [[open,           Var "K",         Var "Vector_in"]])

cf_flexFIRMod_ADD :: Config
cf_flexFIRMod_ADD =
 (["into_(FU_M,1)","into_(FU_A,0)", "into_(FU_A,1)"],
  [[open,           Var "K",         Var "Vector_in"]])

cf_flexFIR_ADD :: Config
cf_flexFIR_ADD = concatConfig 
                   (subConfig cf_flexFIRStartMod_ADD "FIR3") 
		   (subConfigs cf_flexFIRMod_ADD ["FIR2", "FIR1", "FIR0"])


--Both Addition and Multiplication
cf_flexFIRStartMod_BOTH :: Config
cf_flexFIRStartMod_BOTH =
 (["into_(FU_M,1)", "into_(FU_A,0)", "into_(FU_A,1)"],
  [[Var "Vector_in", Var "K",         Var "Vector_in"]])

cf_flexFIRMod_BOTH :: Config
cf_flexFIRMod_BOTH =
 (["into_(FU_M,1)", "into_(FU_A,0)", "into_(FU_A,1)"],
  [[Var "Vector_in", Var "K",         Var "Vector_in"]])

cf_flexFIR_BOTH :: Config
cf_flexFIR_BOTH = concatConfig 
                   (subConfig cf_flexFIRStartMod_BOTH "FIR3") 
		   (subConfigs cf_flexFIRMod_BOTH ["FIR2", "FIR1", "FIR0"])


--Three-Coefficient FIR

cf_flexFIRMod_FIR3 :: Config
cf_flexFIRMod_FIR3 =
 (["into_(FU_M,1)", "into_(FU_A,0)", "into_(FU_A,1)"],
  [[open,            open,            open]])

cf_flexFIR_FIR3 :: Config
cf_flexFIR_FIR3 = concatConfigs
 [
  subConfig cf_flexFIRStartMod_FIR "FIR3",
  subConfig cf_flexFIRMod_FIR "FIR2",
  subConfig cf_flexFIRMod_FIR "FIR1",
  subConfig cf_flexFIRMod_FIR3 "FIR0"
 ]
-}
